/+
 ' Module          : financial.d 
 ' Version         : v1.0 
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 12.Jul.04 Compiled with dmd v0.95.
 ' Modified Date   : 22.Jul.04 Compiled and Tested with dmd v0.96.
 '                 : 24.Nov.04 Recompiled and Tested with dmd v0.106.
 '                 : 09.Dec.04 Added four more functions, and  
 '                 :           recompiled and tested with dmd v0.109.
 '                 : 14.Apr.05 Recompiled and tested with dmd v0.120.
 '                 : 10.Jun.05 Recompiled and tested with dmd v0.126.
 '                 : 28.Jun.05 Added vdb(), xnpv(), xirr(), min(), 
 '                 :           and max() functions. Plus changed
 '                 :           the licence to GPL so I could add these
 '                 :           and even more useful financial functions 
 '                 :           from GPL source code to D.
 '                 : 30.Jun.05 Added effect(), nominal(), dollarde(), 
 '                 :           dollarfr() and recompiled and tested 
 '                 :           dmd v0.127.
 '                 : 22.Mar.06 Recompiled and tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 : 03.Jan.08 Recompiled and Tested with dmd v1.025.
 '                 : 20.Mar.08 Recompiled and Tested with dmd v1.028.
 '                 : 20.Jun.08 Recompiled and Tested with dmd v1.031. 
 '                 : 17.Jul.08 Recompiled and Tested with dmd v1.033. 
 '                 :
 ' Requirements    : std.math, std.date, std.string, and std.stdarg;
 '                 : --------------------------------------------------
 '                 :
 ' Other Authors   : J.W. Rider (c) 1992 
 '                 : * potions of code from "Math Routines for Finance 
 '                 :   Calculations in Turbo Pascal" used and modified.
 '                 :
 '                 : Jukka-Pekka Iivonen (jiivonen@hutcs.cs.hut.fi)
 '                 : Morten Welinder (terra@diku.dk)
 '                 : Vladimir Vuksan (vuksan@veus.hr)
 '                 : Andreas J. Guelzow (aguelzow@taliesin.ca)
 '                 : * potions of code from fn-financial.c used and modified
 '                 :
 '                 : Enrique Garcia M. (c) 2003-2004 (erarcia@egm.as)
 '                 : * potions of code from finance_class.php used and modified
 '                 :
 ' License         : GPL (GNU General Public License)               
 '                 : ---------------------------------------------------------------------
 '                 : This program is free software; you can redistribute it and/or modify
 '                 : it under the terms of the GNU General Public License as published by
 '                 : the Free Software Foundation; either version 2 of the License, or
 '                 : (at your option) any later version.
 ' ---------------------------------------------------------------------------------------
 '  
 ' To Compile for a Unittest:
 ' dmd financial.d -debug=financial -unittest
 '
 ' To Compiled with a Project: 
 ' dmd MyProject.d financial.d
 '
 ' Public Functions:
 ' -----------------
 ' 01) real ddb( in real, in real, in real, in real, in real = 2.0 )
 ' 02) real nper( in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 03) real fv( in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 04) real pmt( in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 05) real ipmt( in real, in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 06) real ppmt( in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 07) real pv( in real, in real, in real, in real = 0.0, in uint = 0 )
 ' 08) real rate( in real, in real, in real, in real = 0.0, in uint = 0, in real = 0.1 )
 ' 09) real sln( in real, in real, in real )
 ' 10) real syd( in real, in real, in real, in real )
 ' 11) real irr( in real[], in real = 0.1 )
 ' 12) real mirr( in real[], in real, in real )
 ' 13) real npv( in real, ... )
 ' 14) real db( in real, in real, in real, in real, uint = 12 )
 ' 15) real ispmt( in real, in real, in real, in real )
 ' 16) real vdb( in real, in real, in real, in real, in real, in real = 2.0, bool = false )
 ' 17) real xnpv( in real, in real[], in d_time[] )
 ' 18) real xirr( in real[], in d_time[], in real = 0.1 )
 ' 19) real effect( in real, in real )
 ' 20) real nominal( in real, in real )
 ' 21) real dollarde( in real, in uint )
 ' 22) real dollarfe( in real, in uint )
 '
 ' Other Needed Functions:
 ' 1) long min( in long, in long )
 ' 2) long max( in long, in long )
 '
 ' -----------------------
 ' Private Functions:
 ' -----------------------
 ' 01) int mfeq( in real, in real, in real )
 ' 
 +/
module financial;

debug( financial ) private import scomp = std.compiler;
private import std.stdio;  // sprintf() and writefln()
private import std.math;   // for pow(), log10(), fabs(), floor(), ceil()
private import std.date;   // for parse() 
private import std.string;
private import std.stdarg; // for argument gathering
private import dateutils;  // getDateAdd(), getDatePart(), getLastDayOfMonth()

const real FINANCIAL_ACCURACY          = 1.0e-6;
const int  FINANCIAL_MAX_ITERATIONS    = 100;
const int  FINANCIAL_BASIS_MSRB_30_360 = 0;
const int  FINANCIAL_BASIS_ACT_ACT     = 1;
const int  FINANCIAL_BASIS_ACT_360     = 2;
const int  FINANCIAL_BASIS_ACT_365     = 3;
const int  FINANCIAL_BASIS_30E_360     = 4;
const int  FINANCIAL_BASIS_30Ep_360    = 5;

long min(in long x, in long y)
{ if (x > y) return y; else return x; }

long max(in long x, in long y)
{ if (x > y) return x; else return y; }

/+
 ' DDB - Double Declining Balance.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: ddb( Cost, Salvage, Life, Period, [Factor] ) 
 '
 ' 1) Cost is the initial cost of the asset. 
 ' 2) Salvage is the value at the end of the depreciation. 
 ' 3) Life is the number of periods over which the asset is being depreciated. 
 ' 4) Period is the period for which you want to calculate the depreciation. (a positive, whole number) 
 ' 5) Factor is the rate at which the balance declines
 '
 ' Having the "factor" available can be handy. Here's why. Double declining balance 
 ' is sometimes called 200 percent (rFactor = 2 ) declining balance. Note that a factor of 2 produces 
 ' the standard DDB. There is a variation of DDB called 150 percent (rFactor = 1.5 ) declining balance 
 ' that uses 150 percent of the straight-line annual percentage rate.
 '
 +/
real ddb
( 
    in real rCost, 
    in real rSalvage, 
    in real rLife, 
    in real rPeriod,
    in real rFactor = 2.0
)
{
    real rx  = 0;
    int  uin = 0;

    while (rPeriod > uin)
    {
        rx = (rFactor * rCost) / rLife;

        if (rCost - rx < rSalvage) 
            rx = rCost - rSalvage;

        if (rx < 0) 
            rx = 0;

        rCost = rCost - rx; 
        uin++; 
    }
    
    return rx;
    
} // end ddb()

unittest
{
    debug( financial ) writefln( "financial.ddb( in real, in real, in real, in real, in real = 2.0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // $1.32
    iMatchs = mfeq( ddb( 2400, 300, 3650, 1 ), 1.31506849315068, fPrecision );
    assert( iMatchs == 1 );  

    // $40.00
    iMatchs = mfeq( ddb( 2400, 300, 120, 1, 2 ), 40, fPrecision );
    assert( iMatchs == 1 );
    
    // $480.00
    iMatchs = mfeq( ddb( 2400, 300, 10, 1, 2 ), 480, fPrecision );
    assert( iMatchs == 1 );

    // $306.00
    iMatchs = mfeq( ddb( 2400, 300, 10, 2, 1.5 ), 306, fPrecision );
    assert( iMatchs == 1 );
    
    // $22.12 ( 22.1225472000002 )
    iMatchs = mfeq( ddb( 2400, 300, 10, 10 ), 22.1225472000002, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' NPer - Number of Periods for an Investment.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: NPer( Rate, Pmt, PV, [FV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) Pmt is the payment made each period. 
 ' 3) PV is the present value. 
 ' 4) FV is the future value. 
 ' 5) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Returns the number of periods for an investment based on periodic,
 ' constant payments and a constant interest rate.
 '
 +/
real nper  
( 
    in real rRate, 
    in real rPmt, 
    in real rPV, 
    in real rFV = 0.0, 
    in uint uiType = 0
)
{
    real rf;

    rf = rPmt * ((rRate * uiType) + 1);
    
    if (std.math.fabs(rRate ) > 1e-6)
        return std.math.log10((rf - rRate * rFV) / (rPV * rRate + rf)) / std.math.log10(1 + rRate);
    else
        return -(rFV + rPV) / (rPV * rRate + rf); 
       
} // end nper()

unittest
{
    debug( financial ) writefln( "financial.nper( in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // 60 ( 59.6738656742946 )
    iMatchs = mfeq( nper( 0.01, -100, -1000, 10000, 1 ), 59.6738656742946, fPrecision );
    assert( iMatchs == 1 );  

    // 60 ( 60.0821228537617 )
    iMatchs = mfeq( nper( 0.01, -100, -1000, 10000 ), 60.0821228537617, fPrecision );
    assert( iMatchs == 1 );
    
    // 11 ( 10.5886444594232 )
    iMatchs = mfeq( nper( 0.01, -100, 1000 ), 10.5886444594232, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' FV - Future Value.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: fv( Rate, NPer, Pmt, [PV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) NPer is total number of payment periods in an annuity. 
 ' 3) Pmt is the payment made each period. 
 ' 4) PV is the present value. 
 ' 5) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Returns the future value of an investment based on periodic,
 ' constant payments and a constant interest rate.
 '
 +/
real fv
( 
    in real rRate, 
    in real rNPer, 
    in real rPmt, 
    in real rPV = 0.0, 
    in uint uiType = 0
)
{
    real rf;

    rf = std.math.exp(rNPer * (log(1 + rRate)));
    
    if (std.math.fabs(rRate) < 1e-6)
        return -rPmt * rNPer * ((rNPer - 1) * rRate / 2 + 1) * (rRate * uiType + 1) - rPV * rf;
    else
        return rPmt * (1 - rf) * (1 / rRate + uiType) - rPV * rf;
        
} // end fv()

unittest
{
    debug( financial ) writefln( "financial.fv( in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // $2,581.40 ( 2581.40337406013 )
    iMatchs = mfeq( fv( 0.005, 10, -200, -500, 1 ), 2581.40337406013, fPrecision );
    assert( iMatchs == 1 );  

    // $12,682.50 ( 12682.503013197 ) 
    iMatchs = mfeq( fv( 0.01, 12, -1000 ), 12682.503013197, fPrecision );
    assert( iMatchs == 1 );
    
    // $82,846.25 ( 82846.2463719006 )
    iMatchs = mfeq( fv( 0.11 / 12, 35, -2000, 0, 1 ), 82846.2463719006, fPrecision );
    assert( iMatchs == 1 );

    // $2,301.40 ( 2301.40183034091 )
    iMatchs = mfeq( fv( 0.005, 12, -100, -1000, 1 ), 2301.40183034091, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' PMT - Payment.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: pmt( Rate, NPer, PV, [FV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) NPer is total number of payment periods in an annuity. 
 ' 3) PV is the present value. 
 ' 4) FV is the future value.  
 ' 5) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Calculates the payment for a loan based on constant payments 
 ' and a constant interest rate.
 '
 +/
real pmt 
( 
    in real rRate, 
    in real rNPer, 
    in real rPV, 
    in real rFV = 0.0, 
    in uint uiType = 0
)
{
    real rf;

    rf = std.math.exp(rNPer * std.math.log(1 + rRate));

    return (rFV + rPV * rf) * rRate / ((1 + rRate * uiType) * (1 - rf)); 
    
} // end pmt()

unittest
{
    debug( financial ) writefln( "financial.pmt( in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // -$1,037.03 ( -1037.03208935916 )
    iMatchs = mfeq( pmt( 0.08 / 12, 10, 10000 ), -1037.03208935916, fPrecision );
    assert( iMatchs == 1 );  

    // -$1,030.16 ( -1030.16432717798 )
    iMatchs = mfeq( pmt( 0.08 / 12, 10, 10000, 0, 1 ), -1030.16432717798, fPrecision );
    assert( iMatchs == 1 );
    
    // $1,030.20 ( 1030.1989980794 )
    iMatchs = mfeq( pmt( 0.12 / 12, 5, -5000 ), 1030.1989980794, fPrecision );
    assert( iMatchs == 1 );

    // -$129.08 ( -129.081160867995 )
    iMatchs = mfeq( pmt( 0.06 / 12, 18 * 12, 0, 50000 ), -129.081160867995, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' IPMT - Interest Payment.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: ipmt( Rate, Per, NPer, PV, [FV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) Per is the period for which you want to find the interest and must be in the range 1 to MPer.
 ' 3) NPer is total number of payment periods in an annuity. 
 ' 4) PV is the present value. 
 ' 5) FV is the future value.  
 ' 6) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Returns the investment payment for a given period for a investment based on periodic,
 ' constant payments and a constant interest rate.
 '
 +/
real ipmt
(
    in real rRate, 
    in real rPer, 
    in real rNPer, 
    in real rPV, 
    in real rFV = 0.0, 
    in uint uiType = 0
)
{
    return rRate * fv(rRate, rPer - uiType -1, pmt(rRate, rNPer, rPV, rFV, uiType), rPV, uiType);
    
} // end ipmt()

unittest
{
    debug( financial ) writefln( "financial.ipmt( in real, in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // -$66.67
    iMatchs = mfeq( ipmt( 0.01 / 12, 1, 36, 80000 ), -66.6666666666667, fPrecision );
    assert( iMatchs == 1 );  

    // -$292.45
    iMatchs = mfeq( ipmt( 0.1, 3, 3, 8000 ), -292.447129909366, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' PPMT - Principal Payment.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: ppmt( Rate, Per, NPer, PV, [FV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) Per is the period for which you want to find the interest and must be in the range 1 to MPer.
 ' 3) NPer is total number of payment periods in an annuity. 
 ' 4) PV is the present value. 
 ' 5) FV is the future value.  
 ' 6) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Returns the payment on the principal for a given period for a investment 
 ' based on periodic, constant payments and a constant interest rate.
 '
 +/
real ppmt
( 
    in real rRate, 
    in real rPer, 
    in real rNPer, 
    in real rPV, 
    in real rFV = 0.0, 
    in uint uiType = 0
)
{
    real rf;

    rf = pmt(rRate, rNPer, rPV, rFV, uiType);
    
    return rf - rRate * fv(rRate, rPer - uiType - 1, rf, rPV, uiType);
    
} // end ppmt()

unittest
{
    debug( financial ) writefln( "financial.ppmt( in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // -$75.62
    iMatchs = mfeq( ppmt( 0.1 / 12, 1, 24, 2000 ), -75.6231860083666, fPrecision );
    assert( iMatchs == 1 );  

    // -$27,598.05
    iMatchs = mfeq( ppmt( 0.08, 10, 10, 200000 ), -27598.0534624214 , fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' PV - Present Value.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: pv( Rate, NPer, Pmt, [FV], [Type] )
 '
 ' 1) Rate is the interest rate per period. 
 ' 2) NPer is total number of payment periods in an annuity. 
 ' 3) Pmt is the payment made each period and cannot change over the life of the annuity. 
 ' 4) FV is the future value.  
 ' 5) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period.
 '
 ' Returns the present value of an investment.
 '
 +/
real pv  
( 
    in real rRate, 
    in real rNPer, 
    in real rPmt, 
    in real rFV = 0.0, 
    in uint uiType = 0
)
{
    real rf;

    if (std.math.fabs(rRate) > 1e-6)
    {
        rf = std.math.exp(rNPer * log(1 + rRate)); 
        return (rPmt * (1 / rRate + uiType) * (1 - rf) - rFV) / rf; 
    }  
    else
        return -(rPmt * (1 + rRate * uiType) * rNPer + rFV) / (1 + rNPer * rRate);
        
} // end pv()

unittest
{
    debug( financial ) writefln( "financial.pv( in real, in real, in real, in real = 0.0, in uint = 0 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // -$59,777.15
    iMatchs = mfeq( pv( 0.08 / 12, 12 * 20, 500 ), -59777.1458511878, fPrecision );
    assert( iMatchs == 1 );  
}

/+
 ' RATE - Interest Rate.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: rate( NPer, Pmt, PV, [FV], [Type], [Guess] )
 '
 ' 1) NPer is total number of payment periods in an annuity. 
 ' 2) Pmt is the payment made each period.   
 ' 3) PV is the present value. 
 ' 4) FV is the future value.  
 ' 5) Type is the rate at which the balance declines.
 '    a) When 0 the payments are due at the end of the period.
 '    b) When 1 the payments are due at the beginning of the period. 
 ' 6) Guess is a number that you guess is close to the result of Rate. 
 '    (0.1 or 10 percent assumed if guess is omitted)
 '
 ' Returns the investment rate per period in an annuity.
 '
 +/ 
real rate 
( 
    in real rNPer, 
    in real rPmt, 
    in real rPV, 
    in real rFV = 0.0, 
    in uint uiType = 0,
    in real rGuess = 0.1
)
{
    real rRate;
    real rx0;
    real rx1;
    real ry0;
    real ry1;

    real y()
    {
        real rf;
  
        if (std.math.fabs(rRate) < 1e-6)
            return rPV * (1 + rNPer * rRate) + rPmt * (1 + rRate * uiType) * rNPer + rFV;
        else 
        {
            rf = std.math.exp(rNPer * log(1 + rRate));
            return rPV * rf + rPmt * (1 / rRate + uiType) * (rf - 1) + rFV; 
        }
    }

    /+
     ' JWR: There are two fundamental problems with solutions by successive
     ' approximation.  One is figuring out where you want to start; the
     ' other is figuring out where you want to stop.  If you don't set them
     ' right, then your solution will approximate successively forever.
     ' This is my guess, but there is no guarantee that the solution will
     ' even exist, much less converge. 
     +/

    rRate = rGuess; 
    ry0   = rPV + rPmt * rNPer + rFV; 
    rx0   = rRate;
    rRate = exp( 1 / rNPer ) - 1; 
    ry1   = y(); 
    rx1   = rRate;
    
    // find root by secant method 
    while (std.math.fabs(ry0 - ry1) > 1e-6) 
    {
       rRate = (ry1 * rx0 - ry0 * rx1) / (ry1 - ry0); 
       rx0 = rx1; 
       rx1 = rRate; 
       ry0 = ry1; 
       ry1 = y(); 
    }
     
    return rRate;
    
} // end rate()

unittest
{
    debug( financial ) writefln("financial.rate( in real, in real, in real, in real = 0.0, in uint = 0, real real = 0.1 ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // 0.77% ( 0.00770147248820 )
    iMatchs = mfeq( rate( 48, -200, 8000 ), 7.70147248820137E-03, fPrecision );
    assert( iMatchs == 1 );  
}

/+
 ' SLN - Straight-Line Depreciation.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: sln( Cost, Salvage, Life ) 
 '
 ' 1) Cost is the initial cost of the asset. 
 ' 2) Salvage is the value at the end of the depreciation. 
 ' 3) Life is the number of periods over which the asset is being depreciated. 
 '
 ' Calculates the straight-line depreciation of an asset for a single period.
 '
 +/
real sln   
( 
    in real rCost, 
    in real rSalvage, 
    in real rLife
)
{
    return (rCost - rSalvage) / rLife; 
    
} // end sln()

unittest
{
    debug( financial ) writefln( "financial.sln( in real, in real, in real ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // $2,250.00
    iMatchs = mfeq( sln( 30000, 7500, 10 ), 2250, fPrecision );
    assert( iMatchs == 1 );  
}

/+
 ' SYD - Sum-of-Years' digits Depreciation
 ' -------------------------------------------------------------------------
 ' Excel/VBA: syd( Cost, Salvage, Life, Period ) 
 '
 ' 1) Cost is the initial cost of the asset. 
 ' 2) Salvage is the value at the end of the depreciation. 
 ' 3) Life is the number of periods over which the asset is being depreciated. 
 ' 4) Period is the period for which you want to calculate the depreciation. (a positive, whole number) 
 '
 ' Calculates the sum-of-years' digits depreciation on an asset for a specified period.
 '
 +/ 
real syd   
( 
    in real rCost, 
    in real rSalvage, 
    in real rLife, 
    in real rPeriod
)
{
    return 2 * (rCost - rSalvage) * (rLife - rPeriod + 1) / (rLife * (rLife + 1)); 
    
} // end syd()

unittest
{
    debug( financial ) writefln( "financial.syd( in real, in real, in real, in real ).unittest"c );

    const float fPrecision = 0.0000001;
    int iMatchs = 0;

    // $4090.91
    iMatchs = mfeq( syd( 30000, 7500, 10, 1 ), 4090.90909090909, fPrecision );
    assert( iMatchs == 1 );  

    // $409.09
    iMatchs = mfeq( syd( 30000, 7500, 10, 10 ), 409.090909090909, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' IRR - Internal Rate-of-Return
 ' -------------------------------------------------------------------------
 ' Excel/VBA: irr( Values(), [Guess] ) 
 '
 ' 1) Values must contain at least one positive and one negative value each
 '    to calculate the internal rate of return. 
 ' 2) Guess is a number that you guess is close to the result of IRR. 
 '    (0.1 or 10 percent assumed if guess is omitted) 
 '
 ' Calculates the internal rate-of-return for a series of cash flows represented
 ' by the numbers in values.
 '
 +/  
real irr
( 
    in real[] rValues, 
    in real   rGuess = 0.1 
)
{
    real rOrate = 0; 
    real rRate  = 0;

    real drate( in real rRate )
    {    
        real rNPv      = 0;
        real rNPvPrime = 0;
        real rValue    = 0;
        int  i         = 0;
        uint uipos     = 0;
  
        rRate = 1 / (1 + rRate);
        
        for ( i = rValues.length - 1; i >= 0; i-- )
        {
            rValue = rValues[i];
            rNPv = rNPv * rRate + rValue;
            rNPvPrime = (rNPvPrime + rValue * i) * rRate;
        }
        
        if (std.math.fabs(rNPvPrime) < 1e-6)
            return rNPv * 1e-6; // a guess 
        else                       
            return rNPv / rNPvPrime; 
            
    } // end drate()
    
    rOrate = rGuess; 
    rRate  = rOrate + drate(rOrate);
    
    // find root by newton-raphson
    while (std.math.fabs(rRate - rOrate) > 1e-6)
    {
        rOrate = rRate; 
        rRate  = rRate + drate(rRate); 
    }
    
    return rRate;
    
} // end iir()

unittest
{
    debug( financial ) writefln( "financial.irr( in real[], in real = 0.1 ).unittest"c );

    const float fPrecision = 0.0000001;
    int  iMatchs = 0;
    real rArr[];
 
    rArr.length = 6;
    rArr[ 0 ] = -70_000; 
    rArr[ 1 ] = 12_000;
    rArr[ 2 ] = 15_000;
    rArr[ 3 ] = 18_000;
    rArr[ 4 ] = 21_000;
    rArr[ 5 ] = 26_000;
    
    // 8.66% ( 0.08663094803625 )
    iMatchs = mfeq( irr( rArr[], 0.1 ), 8.66309480331796E-02, fPrecision );
    assert( iMatchs == 1 );   
    rArr.length = 5;

    // -2.12% ( -0.02124484827341 )
    iMatchs = mfeq( irr( rArr[], 0.1 ), -0.02124484827347, fPrecision );
    assert( iMatchs == 1 );  
    
    rArr.length = 3;
    
    // -44.35% ( -0.443506941334741 )
    iMatchs = mfeq( irr( rArr[], -0.1 ), -0.443506941334741, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' MIRR - Modified Internal Rate of Return. 
 ' -------------------------------------------------------------------------
 ' Excel/VBA: mirr( Values(), Finance_Rate, Reinvest_Rate )
 '
 ' 1) Values must contain at least one positive and one negative value each
 '    to calculate the internal rate of return. 
 ' 2) Finance Rate is the interest rate you pay on the money used in the cash flows.
 ' 3) Reinvest Rate is the interest rate you receive on the cash flows as you reinvest them.
 '
 ' Returns the modified internal rate of return for a series of periodic cash flows.
 +/
real mirr 
( 
    in real[] rValues, 
    in real rFRate, 
    in real rRRate 
)
{
    real rNpvNEG = 0.0;
    real rNpvPOS = 0.0;

    for ( uint uix = 0; uix < rValues.length; uix++ ) 
    {
        if (rValues[uix] >= 0.0)
            rNpvPOS += rValues[uix] / std.math.pow(1 + rRRate, uix);
        else
            rNpvNEG += rValues[uix] / std.math.pow(1 + rFRate, uix);
    }

    // should return a #DIV/0! error value, but currently returning 0;
    if (rNpvNEG == 0 || rNpvPOS == 0 || rRRate <= -1) 
        return real.nan; //0.0;
 
    return std.math.pow((-rNpvPOS * std.math.pow(1 + rRRate, rValues.length)) / (rNpvNEG * (1 + rRRate)), 
           cast(real)(1.0 / (rValues.length - 1))) - 1.0;

} // end mirr()

unittest
{
    debug( financial ) writefln( "financial.mirr( in real[], in real, in real ).unittest"c );

    const float fPrecision = 0.0000001;
    int  iMatchs = 0;
    real rArr[];
 
    rArr.length = 6;
    rArr[ 0 ] = -100_000; 
    rArr[ 1 ] = 100_000;
    rArr[ 2 ] = 100_000;
    rArr[ 3 ] = 100_000;
    rArr[ 4 ] = 100_000;
    rArr[ 5 ] = 100_000;

    // 44.7% ( 0.447417049980915 ) 
    iMatchs = mfeq( mirr( rArr[], 0.10, 0.12 ), 0.447417049980915, fPrecision );
    assert( iMatchs == 1 );

    rArr.length = 6;
    rArr[ 0 ] = -120_000; 
    rArr[ 1 ] = 39_000;
    rArr[ 2 ] = 30_000;
    rArr[ 3 ] = 21_000;
    rArr[ 4 ] = 37_000;
    rArr[ 5 ] = 46_000;

    // 13.0% ( 0.126094130365905 ) 
    iMatchs = mfeq( mirr( rArr[], 0.10, 0.12 ), 0.126094130365905, fPrecision );
    assert( iMatchs == 1 );

    // 13.0% ( 0.134759110828315 ) 
    iMatchs = mfeq( mirr( rArr[], 0.10, 0.14 ), 0.134759110828315, fPrecision );
    assert( iMatchs == 1 );

    rArr.length = 4;
    // -5.0% ( -0.04804465525 ) 
    iMatchs = mfeq( mirr( rArr[], 0.10, 0.12 ), -0.04804465525, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' NPV - Net Present Value. 
 ' -------------------------------------------------------------------------
 ' Excel/VBA: npv( Rate, Value1, Value2, ... )
 '
 ' 1) Rate is the rate of discount over the length of one period. 
 ' 2) Value1, Value2, ...  must be equally spaced in time and occur at the end of each period. 
 '
 ' Returns the net present value of an investment by using a discount rate and a series of 
 ' future payments (negative values) and income (positive values).
 +/
real npv( in real rRate, ... )
{
    real[] rValues;
    real   rSum = 0.0;
    uint   uiy;

    rValues.length = _arguments.length;

    for ( uint uix = 0; uix < _arguments.length; uix++)
    {
        if (_arguments[uix] == typeid(real))
            rValues[uix] = va_arg!(real)(_argptr); 
        else if (_arguments[uix] == typeid(double))
            rValues[uix] = va_arg!(double)(_argptr); 
        else if (_arguments[uix] == typeid(float))
            rValues[uix] = va_arg!(float)(_argptr); 
        else if (_arguments[uix] == typeid(long))
            rValues[uix] = va_arg!(long)(_argptr); 
        else if (_arguments[uix] == typeid(int))
            rValues[uix] = va_arg!(int)(_argptr); 
        else if (_arguments[uix] == typeid(ulong))
            rValues[uix] = va_arg!(ulong)(_argptr); 
        else if (_arguments[uix] == typeid(uint))
            rValues[uix] = va_arg!(uint)(_argptr); 
    }

    uiy  = rValues.length - 1;
    rSum = 0.0;
    
    for ( uint uix = 0; uix < rValues.length; uix++ )
    {
        rSum += (rValues[uix] / std.math.pow(1 + rRate, rValues.length - uiy));
        //writefln( "for-loop:rSum=%10.8f, pow()=%10.8f", rSum, std.math.pow( 1 + rRate, rValues.length - uiy ) );
        uiy--;
    }

    return rSum;

} // end real npv()

unittest
{
    debug( financial ) writefln( "financial.npv( in real, ... ).unittest"c );

    const float fPrecision = 0.0000001;
    int  iMatchs = 0;

    // $1,188.44 ( 1_188.443412335220 ) 
    iMatchs = mfeq( npv( 0.1, -10_000, 3_000, 4_200, 6_800 ), 1_188.443412335220, fPrecision );
    assert( iMatchs == 1 );

    // $1,922.06 ( 41,922.0615549324 - 40,000 ) 
    iMatchs = mfeq( npv( 0.08, 8_000, 9_200, 10_000, 12_000, 14_500 ) - 40_000, 1_922.0615549324, fPrecision );
    assert( iMatchs == 1 );

    // $-3,749.47 ( 36_250.5349129844 - 40_000 = 3_749.46508701557 ) 
    iMatchs = mfeq( npv( 0.08, 8_000, 9_200, 10_000, 12_000, 14_500, -9_000 ) - 40_000, -3_749.46508701557, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' DB - Declining (Fixed) Balance. 
 ' -------------------------------------------------------------------------
 ' Excel/VBA: DB( Cost, Salvage, Life, Period, [Month] )
 '
 ' 1) Cost is the initial cost of the asset. 
 ' 2) Salvage is the value at the end of the depreciation. 
 ' 3) Life is the number of periods over which the asset is being depreciated. 
 ' 4) Period is the period for which you want to calculate the depreciation. (a positive, whole number) 
 ' 5) Month is the number of months in the first year. If month is omitted, it is assumed to be 12.
 ' 
 ' Returns the depreciation of an asset for a specified period using the fixed-declining balance method.
 +/
real db
( 
    in real rCost, 
    in real rSalvage, 
    in real rLife, 
    in real rPeriod, 
    in uint uiMonth = 12U
)
{
    real rRate;
    real rTotal;

    // Need a better error handler here.  
    if (rCost == 0.0L || rLife <= 0.0L || ((rSalvage / rCost) < 0.0L))
        return real.nan; //0.0L;

    rRate = 1.0L - std.math.pow((rSalvage / rCost), (1.0L / rLife));
    rRate *= 1000.0L;
    rRate = std.math.floor(rRate + 0.5L) / 1000.0L;

    rTotal = rCost * rRate * cast(real)uiMonth / 12.0L;

    if (rPeriod == 1.0L)
        return rTotal;

    for ( uint uix = 1U; uix < rLife; uix++ )
    {
       if (uix == rPeriod - 1.0L)
           return (rCost - rTotal) * rRate;
       else
           rTotal += (rCost - rTotal) * rRate;
    }

    return (((rCost - rTotal) * rRate * cast(real)(12U - uiMonth)) / 12.0L );

} // end real db()

unittest
{
    debug( financial ) writefln( "financial.db( in real, in real, in real, in real, uint = 12 ).unittest"c );

    const float fPrecision = 0.0000001;
    int  iMatchs = 0;

    // $186,083.33 Depreciation in first year 0f seven ( 186_083.333333333 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 1, 7 ), 186_083.333333333, fPrecision );
    assert( iMatchs == 1 );

    // $259,639.42 Depreciation in second year ( 259_639.416666667 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 2, 7 ), 259_639.416666667, fPrecision );
    assert( iMatchs == 1 );

    // $176,814.44 Depreciation in third year ( 176_814.4427500 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 3, 7 ), 176_814.4427500, fPrecision );
    assert( iMatchs == 1 );

    // $120,410.64 Depreciation in fourth year ( 120_410.63551275 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 4, 7 ), 120_410.63551275, fPrecision );
    assert( iMatchs == 1 );

    // $81,999.64 Depreciation in fifth year ( 81_999.6427841827 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 5, 7 ), 81_999.6427841827, fPrecision );
    assert( iMatchs == 1 );

    // $55,841.76 Depreciation in sixth year ( 55_841.7567360284 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 6, 7 ), 55_841.7567360284, fPrecision );
    assert( iMatchs == 1 );

    // $15,845.10 Depreciation in seventh year, with only 5 months calculated ( 15_845.0984738481 ) 
    iMatchs = mfeq( db( 1_000_000, 100_000, 6, 7, 7 ), 15_845.0984738481, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' ISPMT - Interest paid during a Specific period of an investment.
 ' -------------------------------------------------------------------------
 ' Excel/VBA: ispmt( Rate, Per, NPer, PV )
 '
 ' 1) Rate is the interest rate for the investment.
 ' 2) Per is the period for which you want to find the interest, and must be between 1 and NPer.
 ' 3) NPer is the total number of payment periods for the investment.
 ' 4) PV is the present value of the investment. For a loan, PV is the loan amount.
 ' 
 ' Returns the interest paid during a specific period of an investment.
 ' -------------------------------------------------------------------------
 ' Remarks: 
 '    Make sure that you are consistent about the units you use for specifying rate and nper. 
 '    If you make monthly payments on a four-year loan at an annual interest rate of 12 percent, 
 '    use 12%/12 for rate and 4*12 for nper. If you make annual payments on the same loan, use 
 '    12% for rate and 4 for nper. 
 '
 '    For all the arguments, the cash you pay out, such as deposits to savings or other withdrawals, 
 '    is represented by negative numbers; the cash you receive, such as dividend checks and other 
 '    deposits, is represented by positive numbers. 
 +/
real ispmt
( 
    in real rRate, 
    in real rPer, 
    in real rNPer, 
    in real rPV
)
{
    real rValue;

    // Need a better error handler here. 
    if (rPer < 1 || rPer > rNPer)
        return real.nan; //0.0;

    rValue = -rPV * rRate;

    return (rValue - (rValue / rNPer * rPer));

} // end real ispmt()

unittest
{
    debug( financial ) writefln( "financial.ispmt( in real, in real, in real, in real ).unittest"c );

    const float fPrecision = 0.0000001;
    int  iMatchs = 0;

    // -$64,814.81 Interest paid for the first monthly payment of a loan ( -64_814.8148148148 ) 
    iMatchs = mfeq( ispmt( 0.10 / 12, 1, 3 * 12, 8_000_000 ), -64_814.8148148148, fPrecision );
    assert( iMatchs == 1 );

    // -$533,333.33 Interest paid in the first year of a loan ( -533_333.333333333 ) 
    iMatchs = mfeq( ispmt( 0.10, 1, 3, 8_000_000 ), -533_333.333333333, fPrecision );
    assert( iMatchs == 1 );
}

/+
 ' VDB - Variable Declining Balance. 
 ' -------------------------------------------------------------------------------------
 ' Excel/VBA: DB( Cost, Salvage, Life, Start_Period, End_Period, [Factor], [No_Switch] )
 '
 ' 1) Cost is the initial cost of the asset. 
 ' 2) Salvage is the value at the end of the depreciation. 
 ' 3) Life is the number of periods over which the asset is being depreciated. 
 ' 4) Start_Period is the starting period for which you want to calculate the depreciation. 
 '    Start_period must use the same units as life. (a positive, whole number) 
 ' 5) End_Period is the ending period for which you want to calculate the depreciation. 
 '    End_period must use the same units as life. (a positive, whole number) 
 ' 6) Factor is the rate at which the balance declines.
 '    * Having the "factor" available can be handy. Here's why. Double declining balance 
 '      is sometimes called 200 percent (rFactor = 2 ) declining balance. Note that a factor 
 '      of 2 produces the standard DDB. There is a variation of DDB called 150 percent 
 '      (rFactor = 1.5 ) declining balance that uses 150 percent of the straight-line annual 
 '      percentage rate.
 ' 7) No_Switch is a logical value specifying whether to switch to straight-line depreciation 
 '    when depreciation is greater than the declining balance calculation. 
 '    * If no_switch is TRUE, Microsoft Excel does not switch to straight-line depreciation even 
 '      when the depreciation is greater than the declining balance calculation. 
 '    * If no_switch is FALSE or omitted, Excel switches to straight-line depreciation when 
 '      depreciation is greater than the declining balance calculation. 
 '
 ' Returns the depreciation of an asset for a specified period using the fixed-declining balance method.
 +/
real vdb
(
    in real rCost, 
    in real rSalvage, 
    in real rLife, 
    in real rStart_Period, 
    in real rEnd_Period, 
    in real rFactor    = 2.0L, 
    in bool bNo_Switch = false
)
{
    int  iLifeAccum = 0;
    real rTerm      = 0.0L;
    real rVDB       = 0.0L;
    int  iInitStart = 0;
    int  iInitEnd   = 0;
    int  iSetStart  = 0;
    int  iSetEnd    = 0;
    real rPeriod    = 0.0L;

    if ((rStart_Period < 0.0L) ||
        (rEnd_Period < rStart_Period) ||
        (rEnd_Period > cast(int)rLife) ||
        (rCost < 0.0L) || (rSalvage > rCost) ||
        (rFactor <= 0.0L))
        return real.nan; //null

    real ivdb
    (
        in real rCost, 
        in real rSalvage, 
        in real rLife, 
        in real rLifex, 
        in real rPeriod, 
        in real rFactor
    )
    {
        real rVDB       = 0.0L;
        real rInitEnd   = std.math.ceil(rPeriod);
        int  iEndPeriod = cast(int)rInitEnd;
        real rRCost     = rCost - rSalvage;
        bool bAtTerm    = false;
        real rDDB       = 0.0L;
        real rTerm      = 0.0L;
        real rLiab      = 0.0L;

        for (int iPeriod = 1; iPeriod <= iEndPeriod; iPeriod++) 
        {
            if (!bAtTerm) 
            {
                rDDB  = ddb(rCost, rSalvage, rLife, iPeriod, rFactor);
                rLiab = rRCost / (rLifex - cast(real)(iPeriod - 1));

                if (rLiab > rDDB) 
                {
                    rTerm   = rLiab;
                    bAtTerm = true;
                } 
                else 
                {
                    rTerm   = rDDB;
                    rRCost -= rDDB;
                }
            } 
            else
                rTerm = rLiab;

            if (iPeriod == iEndPeriod)
                rTerm *= (rPeriod + 1.0L - rInitEnd);
            
            rVDB += rTerm;
        }
    
        return rVDB;

    } // end ivdb()

    rVDB       = 0.0L;
    iInitStart = cast(int)std.math.floor(rStart_Period);
    iInitEnd   = cast(int)std.math.ceil(rEnd_Period);

    if (bNo_Switch) 
    {
        iSetStart = iInitStart;
        iSetEnd   = iInitEnd;

        for (int iCurrPeriod = iSetStart + 1; iCurrPeriod <= iSetEnd; iCurrPeriod++) 
        {
            rTerm = ddb(rCost, rSalvage, rLife, iCurrPeriod, rFactor);

            if (iCurrPeriod == iSetStart + 1)
                rTerm *= (min(cast(long)rEnd_Period, iInitStart + 1) - cast(long)rStart_Period);
            else if (iCurrPeriod == iSetEnd)
                rTerm *= (cast(long)rEnd_Period + 1 - iInitEnd);
                
            rVDB += rTerm;
        }
    } 
    else 
    {
        iLifeAccum = cast(int)rLife;

        if (cast(int)rStart_Period != iInitStart)
            if (rFactor > 1.0L) 
            {
                if (cast(int)rStart_Period >= (cast(int)rLife / 2)) 
                {
                    rPeriod        = rStart_Period - (rLife / 2);
                    rStart_Period  = rLife / 2;
                    rEnd_Period   -= rPeriod;
                    iLifeAccum    += 1;
                }
            }

        rCost -= ivdb(rCost, rSalvage, rLife, iLifeAccum, rStart_Period, rFactor);
        rVDB   = ivdb(rCost, rSalvage, rLife, rLife - rStart_Period, rEnd_Period - rStart_Period, rFactor);
    }

    return rVDB;

} // end vdb()

unittest
{
    debug( financial ) writefln( "financial.vdb(in real, in real, in real, in real, in real, in real = 2.0, bool = false).unittest"c );
    const float fPrecision = 0.0001;
    int  iMatch  = 0;
    real rResult = 0.0L;

    rResult = vdb(2400, 300, 120, 0, 1);
    iMatch  = mfeq(rResult, 40.0000, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(10000, 500, 120, 2, 4);
    iMatch  = mfeq(rResult, 319.6289, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(10000, 1000, 9, 3, 4);
    iMatch  = mfeq(rResult, 1045.5723, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(1000, 1000, 9, 3, 3);
    iMatch  = mfeq(rResult, 0.0000, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 10, 0, 1);
    iMatch  = mfeq(rResult, 480.0000, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 120, 6, 18);
    iMatch  = mfeq(rResult, 396.3061, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 10, 0, 0.875, 1.5);
    iMatch  = mfeq(rResult, 315.0000, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 3650, 0, 1);
    iMatch  = mfeq(rResult, 1.3151, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 120, 6, 18);
    iMatch  = mfeq(rResult, 396.3061, fPrecision);
    assert(iMatch == 1);

    rResult = vdb(2400, 300, 120, 6, 18, 1.5);
    iMatch  = mfeq(rResult, 311.8089, fPrecision);
    assert(iMatch == 1);
}

/+
 ' XNPV - calculates the Net Present Value of an investment. 
 ' -------------------------------------------------------------------------
 ' Excel (add-in) : xnpv( Rate, Payments[], Dates[] )
 '
 ' 1) Rate is the rate of discount over the length of one period. 
 '
 ' 2) Payments[] is a series of cash flows that corresponds to a schedule of payments in dates. 
 '    The first payment is optional and corresponds to a cost or payment that occurs at the 
 '    beginning of the investment. If the first value is a cost or payment, it must be a negative 
 '    value. All succeeding payments are discounted based on a 365-day year. The series of values 
 '    must contain at least one positive and one negative value. (Payments must be equally spaced 
 '    in time and occur at the end of each period.)
 '
 ' 3) Dates[] is a schedule of payment dates that corresponds to the cash flow payments. The first
 '    payment date indicates the beginning of the schedule of payments. All other  dates must be 
 '    later than this date, but they may occur in any order.
 '
 ' Returns the net present value for a schedule of cash flows that's
 ' given in dates array is not necessarily periodic. The first date 
 ' indicates the beginning of the payment schedule. rate is the interest 
 ' rate and values are the payments.
 '
 ' To calculate the net present value for a series of cash flows that is 
 ' periodic, use the NPV function instead.
 +/
real xnpv(in real rRate, in real[] rPayments, in d_time[] dtDates)
{
    if (rPayments.length != dtDates.length) 
        return real.nan; //null;

    real rXNPV  = 0.0L;
    real rValue = 0;
    long lDays  = 0L;
    
    for ( int i = 0; i < rPayments.length; i++ )
    {
        lDays  = cast(long)std.math.ceil((dtDates[i] - dtDates[0]) / std.date.TicksPerDay);
        rValue = (lDays / 365.0L);
        rXNPV  = rXNPV + rPayments[i] / std.math.pow(1.0L + rRate, rValue);
    }
    return (std.math.isfinite(rXNPV) ? rXNPV : real.nan); // : null);

} // end xnpv()

/+
 ' NOTE: That the Date string parameters can be formatted as
 '       "MM/DD/YYYY", "YYYY/MM/DD"
 '       "DD MMM YYYY", MMM DD YYYY", "YYYY MMM DD",
 '       "D M. YYYY", "M. D YYYY", YYYY M. D".
 +/
real xnpv(in real rRate, in real[] rPayments, in char[][] sDates)
{
    d_time[] dtDates;
    
    dtDates.length = sDates.length;

    for ( int i = 0; i < sDates.length; i++ )
        dtDates[i] = std.date.parse(sDates[i]);

    return xnpv(rRate, rPayments[], dtDates[]);
}

unittest
{
    debug( financial ) writefln( "financial.xnpv(in real, in real[], in d_time[] ).unittest"c );
    const float fPrecision = 0.0001;

    int      iMatch  = 0;
    real     rResult = 0.0L;
    real[]   rCashFlow;
    d_time[] dtCashFlowDates;

    rCashFlow.length = 5;
    dtCashFlowDates.length = 5;

    // 1st Test
    rCashFlow[0] = -12000.00;
    rCashFlow[1] = 3000.00;
    rCashFlow[2] = 4150.00;
    rCashFlow[3] = 3300.00;
    rCashFlow[4] = 3150.00;

    dtCashFlowDates[0] = std.date.parse("01/10/1991"c);
    dtCashFlowDates[1] = std.date.parse("04/01/1991"c);
    dtCashFlowDates[2] = std.date.parse("09/30/1991"c);
    dtCashFlowDates[3] = std.date.parse("01/10/1992"c);
    dtCashFlowDates[4] = std.date.parse("04/01/1992"c);

    // aws = 705.334197 or $705.33
    rResult = xnpv(0.09, rCashFlow[], dtCashFlowDates[]);
    iMatch  = mfeq(rResult, 705.334197, fPrecision);
    assert(iMatch == 1);
    //writefln("1st xnpv=%f", rResult);

    // 2nd Test
    rCashFlow[0] = -10000.00;
    rCashFlow[1] = 2750.00;
    rCashFlow[2] = 4250.00;
    rCashFlow[3] = 3250.00;
    rCashFlow[4] = 2750.00;

    dtCashFlowDates[0] = std.date.parse("01/01/2008"c);
    dtCashFlowDates[1] = std.date.parse("03/01/2008"c);
    dtCashFlowDates[2] = std.date.parse("10/30/2008"c);
    dtCashFlowDates[3] = std.date.parse("02/15/2009"c);
    dtCashFlowDates[4] = std.date.parse("04/01/2009"c);

    // aws 2086.647602 or $2086.65 
    rResult = xnpv(0.09L, rCashFlow[], dtCashFlowDates[]);
    iMatch  = mfeq(rResult, 2086.647602, fPrecision);
    assert(iMatch == 1);
    //writefln("2nd xnpv=%f", rResult);    
}

/+
 ' XIRR - Interest Rate for a series of cash flow. 
 ' -------------------------------------------------------------------------
 ' Excel (add-in) : xirr( Payments[], Dates[], [Guess] )
 '
 ' 1) Payments[] is a series of cash flows that corresponds to a schedule of payments in dates. 
 '    The first payment is optional and corresponds to a cost or payment that occurs at the 
 '    beginning of the investment. If the first value is a cost or payment, it must be a negative 
 '    value. All succeeding payments are discounted based on a 365-day year. The series of values 
 '    must contain at least one positive and one negative value. (Payments must be equally spaced 
 '    in time and occur at the end of each period.)
 '
 ' 2) Dates[] is a schedule of payment dates that corresponds to the cash flow payments. The first
 '    payment date indicates the beginning of the schedule of payments. All other  dates must be 
 '    later than this date, but they may occur in any order.
 '
 ' 3) Guess is a number that you guess is close to the result of Rate. 
 '    (0.1 or 10 percent assumed if guess is omitted)
 '
 ' Returns the internal rate of return for a schedule of cash flows
 ' that is not necessarily periodic. To calculate the internal rate
 ' of return for a series of periodic cash flows, use the IRR function.
 +/
real xirr(in real[] rPayments, in d_time[] dtDates, in real rGuess = 0.1L)
{
    if (rPayments.length != dtDates.length) 
        return real.nan; //null;

    // create an initial bracket, with a root somewhere between bot and top
    real rX1 = 0.0L;
    real rX2 = rGuess;
    real rF1 = xnpv(rX1, rPayments[], dtDates[]);
    real rF2 = xnpv(rX2, rPayments[], dtDates[]);
    real rF0 = 0.0L;
    real rTB = 0.0L;
    real rDX = 0.0L;
    real rX_Mid = 0.0L;
    real rF_Mid = 0.0L;

    for ( int i = 0; i < FINANCIAL_MAX_ITERATIONS; i++ )
    {
        if ((rF1 * rF2) < 0.0L) 
            break;

        if (std.math.fabs(rF1) < std.math.fabs(rF2)) 
            rF1 = xnpv(rX1 += 1.6L * (rX1 - rX2), rPayments[], dtDates[]);
        else 
            rF2 = xnpv(rX2 += 1.6L * (rX2 - rX1), rPayments[], dtDates[]);
    }

    if ((rF1 * rF2) > 0.0L) 
        return real.nan; //null;

    rF0 = xnpv(rX1, rPayments, dtDates);

    if (rF0 < 0.0L) 
    {
        rTB = rX1;
        rDX = rX2 - rX1;
    } 
    else 
    {
        rTB = rX2;
        rDX = rX1 - rX2;
    }

    for ( int i = 0; i < FINANCIAL_MAX_ITERATIONS; i++ )
    {
        rDX *= 0.5L;
        rX_Mid = rTB + rDX;
        rF_Mid = xnpv(rX_Mid, rPayments, dtDates);

        if (rF_Mid <= 0.0L) 
            rTB = rX_Mid;

        if ((std.math.fabs(rF_Mid) < FINANCIAL_ACCURACY) || 
            (std.math.fabs(rDX) < FINANCIAL_ACCURACY)) 
            return rX_Mid;
    }

    return real.nan; //null;

} // end xirr()

real xirr(in real[] rPayments, in char[][] sDates, in real rGuess = 0.1L)
{
    d_time[] dtDates;
    
    dtDates.length = sDates.length;

    for ( int i = 0; i < sDates.length; i++ )
        dtDates[i] = std.date.parse(sDates[i]);

    return xirr(rPayments[], dtDates[], rGuess);
}

unittest
{
    debug( financial ) writefln( "financial.xirr(in real[], in d_time[], in real = 0.1 ).unittest"c );
    const float fPrecision = 0.0001;
    int  iMatch  = 0;
    real rResult = 0.0L;
    real[] rCashFlow;
    d_time[] dtCashFlowDates;

    rCashFlow.length = 5;
    dtCashFlowDates.length = 5;

    // 1st Test
    rCashFlow[0] = -6000.00;
    rCashFlow[1] = 2134.00;
    rCashFlow[2] = 1422.00;
    rCashFlow[3] = 1933.00;
    rCashFlow[4] = 1422.00;

    dtCashFlowDates[0] = std.date.parse("01/15/1999"c);
    dtCashFlowDates[1] = std.date.parse("04/04/1999"c);
    dtCashFlowDates[2] = std.date.parse("05/09/1999"c);
    dtCashFlowDates[3] = std.date.parse("03/12/2000"c);
    dtCashFlowDates[4] = std.date.parse("05/01/2000"c);

    // aws 0.224837691 or 22.48%
    rResult = xirr(rCashFlow[], dtCashFlowDates[]);
    iMatch  = mfeq(rResult, 0.224837, fPrecision);
    assert(iMatch == 1);
    //writefln("1st xirr=%f", rResult);

    // 2nd Test
    rCashFlow[0] = -10000.00;
    rCashFlow[1] = 2500.00;
    rCashFlow[2] = 2000.00;
    rCashFlow[3] = 3000.00;
    rCashFlow[4] = 4000.00;

    dtCashFlowDates[0] = std.date.parse("01/12/2000"c);
    dtCashFlowDates[1] = std.date.parse("02/14/2001"c);
    dtCashFlowDates[2] = std.date.parse("03/03/2001"c);
    dtCashFlowDates[3] = std.date.parse("06/14/2001"c);
    dtCashFlowDates[4] = std.date.parse("12/01/2001"c);

    // aws 0.100598973 or 10.06% 
    rResult = xirr(rCashFlow[], dtCashFlowDates[]);
    iMatch  = mfeq(rResult, 0.100599, fPrecision);
    assert(iMatch == 1);
    //writefln("2nd xirr=%f", rResult);

    // 3rd Test
    rCashFlow[0] = -10000.00;
    rCashFlow[1] = 2750.00;
    rCashFlow[2] = 4250.00;
    rCashFlow[3] = 3250.00;
    rCashFlow[4] = 2750.00;

    dtCashFlowDates[0] = std.date.parse("01/01/2008"c);
    dtCashFlowDates[1] = std.date.parse("03/01/2008"c);
    dtCashFlowDates[2] = std.date.parse("10/30/2008"c);
    dtCashFlowDates[3] = std.date.parse("02/15/2009"c);
    dtCashFlowDates[4] = std.date.parse("04/01/2009"c);

    // aws 0.373362535 or 37.34% 
    rResult = xirr(rCashFlow[], dtCashFlowDates[]);
    iMatch  = mfeq(rResult, 0.373363, fPrecision);
    assert(iMatch == 1);
    //writefln("3rd xirr=%f", rResult);    
}

/+
 ' EFFECT - the effective annual interest rate
 ' ---------------------------------------------------------------------
 ' Excel : effect( Nominal_Rate, NPery )
 '
 ' 1) Nominal_Rate is the nominal annual interest rate.
 ' 2) NPery is number of compounding periods per year.
 ' 
 ' Returns the effective annual interest rate, given the nominal annual
 ' interest rate and the number of compounding periods per year.
 +/
real effect(in real rNominalRate, in real rNumPerYr)
{
    real rEffect = 0.0L;

    if ((rNominalRate <= 0.0L) || (rNumPerYr < 1.0L)) 
        return real.nan; //null;

    rEffect = std.math.pow(1.0L + rNominalRate / rNumPerYr, rNumPerYr) - 1.0L;

    return (std.math.isfinite(rEffect) ? rEffect : real.nan); //0.0L); //null);
}

/+
 ' NOMINAL - the nominal annual interest rate
 ' ---------------------------------------------------------------------
 ' Excel : nominal( Effect_Rate, NPery )
 '
 ' 1) Effect_Rate the effective rate.
 ' 2) NPery is number of compounding periods per year.
 ' 
 ' Returns the nominal annual interest rate, given the effective rate
 ' and the number of compounding periods per year.
 +/
real nominal(in real rEffectRate, in real rNumPerYr)
{
    real rNominal = 0.0L;

    if ((rEffectRate <= 0.0L) || (rNumPerYr < 1.0L)) 
        return real.nan; //null;

    rNominal = rNumPerYr * std.math.pow(1.0L + rEffectRate, 1.0L / rNumPerYr) - rNumPerYr;

    return (std.math.isfinite(rNominal) ? rNominal : real.nan); //0.0L); // null);
}

unittest
{
    const float fPrecision = 0.0001;
    int  iMatch  = 0;
    real rResult = 0.0L;

    debug( financial ) writefln( "financial.effect( in real, in real ).unittest"c );
    rResult = effect(0.9L, 12);
    iMatch  = mfeq(rResult, 1.381779599, fPrecision);
    assert(iMatch == 1);

    debug( financial ) writefln( "financial.nominal( in real, in real ).unittest"c );
    rResult = nominal(1.381779599, 12); 
    iMatch  = mfeq(rResult, 0.9, fPrecision);
    assert(iMatch == 1);
}

/+
 ' DOLLARDE - dollar price expressed as a decimal number
 ' -------------------------------------------------------------
 ' Excel : dollarde( Fractional_Dollar, Fraction )
 '
 ' 1) Fractional_Dollar is a number expressed as a fraction.
 ' 2) Fraction is the integer to use in the denominator of the fraction.
 '
 ' Returns a dollar price expressed as a decimal number.
 '
 ' Note: Converts a dollar price expressed as a fraction into a dollar
 ' price expressed as a decimal number. Use DOLLARDE to convert
 ' fractional dollar numbers, such as securities prices, to decimal
 ' numbers.
 +/
real dollarde( in real rFractionalDollar, in uint uiFraction )
{
    long lFD = cast(long)rFractionalDollar;

    if (uiFraction < 1)
        return real.nan; // null;

    return cast(real)(lFD + (10 * (rFractionalDollar - lFD)) / uiFraction);
}

/+
 ' DOLLARFR - dollar price expressed as a fraction
 ' ----------------------------------------------------------------
 ' Excel : dollarfr( Decimal_Dollar, Fraction )
 '
 ' 1) Decimal_Dollar is a decimal number.
 ' 2) Fraction is the integer to use in the denominator of the fraction.
 '
 ' Returns a dollar price expressed as a fraction.
 '
 ' Note: Converts a dollar price expressed as a decimal number into a
 ' dollar price expressed as a fraction. Use DOLLARFR to convert
 ' decimal numbers to fractional dollar numbers, such as securities
 ' prices.
 +/
real dollarfr( in real rDecimalDollar, in uint uiFraction)
{
    long lDD = cast(long)rDecimalDollar;

    if (uiFraction < 1)
        return real.nan; // null;

    return cast(real)(((rDecimalDollar - lDD) * uiFraction) / 10 + lDD);
}

unittest
{
    const float fPrecision = 0.0001;
    int  iMatch  = 0;
    real rResult = 0.0L;

    debug( financial ) writefln( "financial.dollarde( in real, in uint ).unittest"c );
    rResult = dollarde(0.9876L, 4);
    //writefln("dollarde=%f", rResult);
    iMatch  = mfeq(rResult, 2.469L, fPrecision);
    assert(iMatch == 1);

    debug( financial ) writefln( "financial.dollarfr( in real, in uint ).unittest"c );
    rResult = dollarfr(2.469L, 4); 
    //writefln("dollarfe=%f", rResult);
    iMatch  = mfeq(rResult, 2.1876L, fPrecision);
    assert(iMatch == 1);

    //rResult = dollarfr(-56.45, 9);
    //writefln("rResult=%f", rResult);
}

debug( financial )
{ 
int main()
{ 
	writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("financial.unittest finished!"c); 

    return 0;
}
} // end debug( financial )

/+
 ' Copied this out of std.math because it was set to private, and I wanted to do
 ' unittest compares against the float values to a percision of 0.0000001.
 '
 ' I asked Walter to make this function public in the D User's Forum. 
 '
 ' Walter's reply:
 ' mfreq() is, frankly, a lousy function. I don't want to enshrine it and
 ' thereby have it live forever <g>. A properly written one would interpret the
 ' precision' as the number of bits in the mantissa that should match.
 '
 +/
/+
 ' Simple function to compare two floating point values
 ' to a specified precision.
 ' Returns:
 ' 1  match
 ' 0  nomatch
 +/
private int mfeq( in real rx, in real ry, in real rPrecision )
{
    if ( rx == ry )
        return 1;

    if ( isnan( rx ) )
        return isnan( ry );

    if ( isnan( ry ) )
        return 0;

    return std.math.fabs( rx - ry ) <= rPrecision;
   
} // end int mfeq( in real, in real, in real )

void die( in char[] s )
{
    throw new Exception(s);
}